/* 15-25 November 2023, by Ruud Harmsen
   See rhmidifuncs.h for more.
 */
#include "rhmidifuncs.h"
#include <stdlib.h>
#include <string.h>

static int AllocAndFillDatabuf (
   FILE *fpi, long len, struct rhmidi_eventdata *p_evtdat);

/***********************************************************************/
long *ReadMidiHeadTrackstarts (FILE *fpi,
   int *p_NumTracks, long *p_WhereNumTracks,
   int *p_TicksPerQuarter, int *p_err)
{
   char buf[4+4+2+2+2], *p = buf;
   long hdlen;
   long trklen;
   long midiformat;
   long numtracks;
   long division;
   long *TrackStarts;
   int i;

   if (fread(buf, 1, sizeof buf, fpi) < sizeof buf)
   {
      if (p_err) *p_err = 1;
      return NULL;
   }
   if (strncmp((char *)p, "MThd", 4) != 0)
   {
      if (p_err) *p_err = 2;
      return NULL;
   }
   CnvFmHighByteFirst(p += 4, 4, &hdlen);
   CnvFmHighByteFirst(p += 4, 2, &midiformat);
   if (midiformat != 0 && midiformat != 1)
   {
      if (p_err) *p_err = 3;
      return NULL;
   }
   CnvFmHighByteFirst(p += 2, 2, &numtracks);
   *p_NumTracks = numtracks;

   *p_WhereNumTracks = p - buf;
   CnvFmHighByteFirst(p += 2, 2, &division);
   if (division & 0x8000)
   {
      if (p_err) *p_err = 4;
      return NULL;
   }
   *p_TicksPerQuarter = division;

   if (hdlen > 6)
   {
      /* Skip and ignore extra header bytes in future versions, if any. */
      fseek(fpi, hdlen - 6, SEEK_CUR);
   }
   if (!(TrackStarts = calloc(*p_NumTracks, sizeof TrackStarts[0])))
   {
      if (p_err) *p_err = 5;
      return NULL;
   }

   for (i = 0; i < *p_NumTracks; i++)
   {
      if (fread(buf, 1, 8, fpi) < 8)
      {
         if (p_err) *p_err = 6;
         return NULL;
      }
      if (strncmp((char *)buf, "MTrk", 4) != 0)
      {
         if (p_err) *p_err = 6;
         return NULL;
      }
      CnvFmHighByteFirst(buf + 4, 4, &trklen);
      TrackStarts[i] = ftell(fpi);
      fseek(fpi, trklen, SEEK_CUR);
   }

   return TrackStarts;
}

/***********************************************************************/
void CnvFmHighByteFirst (char *buf, int numbytes, long *p_count)
{
   int i;

   *p_count = 0;
   for (i = numbytes - 1; i >= 0; i--, buf++)
      *p_count += ((unsigned char)(*buf) << (8*i));
}

/***********************************************************************/
void CnvToHighByteFirst (char *buf, int numbytes, long count)
{
   int i;

   for (i = numbytes - 1; i >= 0; i--, buf++)
      *buf = (count & (0xff << (8*i))) >> (8*i);
}

/***********************************************************************/
float CnvFmMidiBend (char *buf)
{
   int lsb = buf[0] & 0x7f;
   int msb = buf[1] & 0x7f;
   float cents;

   cents = (msb << 7) + lsb;
   cents -= 0x2000;
   cents = cents / 0x2000 * 200;

   return cents;
}

/***********************************************************************/
void CnvToMidiBend (char *buf, float cents)
{
   int val14bits, lsb, msb;

   if (cents >= 200) cents =  199.976;
   if (cents < -200) cents = -200;

   cents = cents / 200 * 0x2000;
   cents += 0x2000;

   val14bits = cents + 0.5;
   lsb = val14bits & 0x7f;
   msb = val14bits >> 7;
   buf[0] = lsb; buf[1] = msb;
}

/***********************************************************************/
int CnvFmVariableLengthValue (unsigned char *buf, long *p_Delta_T)
{
   unsigned char *p;
   int bytes = 0;

   for (*p_Delta_T = 0, p = buf; p - buf < 4; p++)
   {
      bytes++;
      *p_Delta_T = (*p_Delta_T | (*p & 0x7f));
      if (!(*p & 0x80))
         break;
      *p_Delta_T = (*p_Delta_T << 7);
   }
   return bytes;
}

/***********************************************************************/
int CnvToVariableLengthValue (unsigned char *buf, long Delta_T)
{
   unsigned char prep[4], *p, *q;
   int bytes = 0;

   for (p = prep; p - prep < 4; p++)
   {
      *p = (Delta_T & 0x7f);
      Delta_T = (Delta_T >> 7);
   }
   for (p = prep + 1; p - prep < 4; p++)
      *p |= 0x80;
   for (p = prep + 4 - 1, q = buf; p > prep; p--)
   {
      if (!bytes && *p != 0x80)
        bytes = p - prep + 1;
      if (bytes)
        *q++ = *p;
  }
   *q = *p;
   if (bytes == 0) bytes = 1;

   return bytes;
}

/***********************************************************************/
int ReadMidiEvent (
   FILE *fpi, struct rhmidi_eventdata *p_evtdat, int TicksPerQuarter)
{
   unsigned char buf[4];
   int retval;
   size_t red;
   long startpos;
   long len;

   fseek(fpi, p_evtdat->TrackFilepos, SEEK_SET);

   if ((red = fread(buf, 1, sizeof buf, fpi)) != sizeof buf)
   {
      p_evtdat->TrackFilepos = ftell(fpi);
      return 1;
   }
   red = CnvFmVariableLengthValue(buf, &p_evtdat->DeltaT);
   if (red < 4)
      /* Reset file position because of variable length encoding,
         that often takes less than 4 bytes. */
      fseek(fpi, red - 4, SEEK_CUR);

   if ((red = fread(buf, 1, 2, fpi)) != 2)
   {
      p_evtdat->TrackFilepos = ftell(fpi);
      return 2;
   }

   if ((buf[0] & 0xF0) == 0xF0)
   {
      /* Meta event (FF), or System Exclusive (F0 and F7).
         Length follows, encoded as variable length.
         Put contents in malloc'ed buffer. */
      p_evtdat->FrstCode = buf[0];
      p_evtdat->ScndCode = buf[1];

      /* End-of-track event */
      if (p_evtdat->FrstCode == 0xff && p_evtdat->ScndCode == 0x2f)
      {
         p_evtdat->EndOfTrack = 1;
         return -1;
      }
      fread(buf, 1, sizeof buf, fpi);
      red = CnvFmVariableLengthValue(buf, &len);

      /* Reset file position because of variable length encoding,
         that often takes less than 4 bytes. */
      fseek(fpi, red - 4, SEEK_CUR);
   }
   else if ((buf[0] & 0x80) == 0x80)
   {
      /* Status code of a channel Voice Message */
      startpos = ftell(fpi) - 1;
      fseek(fpi, startpos, SEEK_SET);

      p_evtdat->StatusCodeLastSeen = buf[0];
      p_evtdat->FrstCode = buf[0] & 0xF0; /* Status code */
      p_evtdat->ScndCode = buf[0] & 0x0F; /* Midi channel */

      len = 2;
      if (p_evtdat->FrstCode == 0xC0 || p_evtdat->FrstCode == 0xD0)
         len = 1;
   }
   else
   {
      /* Running status (code was omitted), so assume value last seen. */
      startpos = ftell(fpi) - 2;
      fseek(fpi, startpos, SEEK_SET);

      p_evtdat->FrstCode = p_evtdat->StatusCodeLastSeen & 0xF0;
      /* Midi channel */
      p_evtdat->ScndCode = p_evtdat->StatusCodeLastSeen & 0x0F;

      len = 2;
      if (p_evtdat->FrstCode == 0xC0 || p_evtdat->FrstCode == 0xD0)
         len = 1;
   }

   retval = AllocAndFillDatabuf(fpi, len, p_evtdat);
   p_evtdat->EventDataLen = len;
   p_evtdat->TrackFilepos = ftell(fpi);
   p_evtdat->SigmaT += p_evtdat->DeltaT;
   p_evtdat->SigmaQ = p_evtdat->SigmaT / (double)TicksPerQuarter;

   return retval;
}

/***********************************************************************/
int ReadMidiEventErrHandling (
   FILE *fpi, char *ProgPath, char *MidiFilename,
   struct rhmidi_eventdata *p_evtdat,
   int TicksPerQuarter, int verbose)
{
   int retval = ReadMidiEvent(fpi, p_evtdat, TicksPerQuarter);

   if (retval > 0)
      fprintf(stderr,
         "%s: In file %s: retval %d\n", ProgPath, MidiFilename, retval);
   switch (retval)
   {
      case 1:
      case 2:
      case 4:
         fprintf(stderr,
            "Premature end of file at byte pos %6ld (dec), %04lx (hex)\n",
            p_evtdat->TrackFilepos, p_evtdat->TrackFilepos);
         p_evtdat->EndOfTrack = 1;
         retval = EOF;
         break;
      case 3:
         fprintf(stderr,
            "%s: In file %s: cannot malloc data buffer. Filepos: %ld\n",
            ProgPath, MidiFilename, p_evtdat->TrackFilepos);
         break;
   }
   if (verbose > 2)
   {
      fprintf(stdout,
         "filepos %4ld %04lx, track %2d: %7.3f qn, 1st-2ndCode %02x-%02x, ",
         p_evtdat->TrackFilepos, p_evtdat->TrackFilepos, p_evtdat->TrackNumber,
         p_evtdat->SigmaQ, p_evtdat->FrstCode, p_evtdat->ScndCode);
      if (p_evtdat->EventData)
      {
         int i;
         fprintf(stdout, "data: ");
         for (i = 0; i < p_evtdat->EventDataLen; i++)
            fprintf(stdout, "%02x ", p_evtdat->EventData[i]);
      }
      fprintf(stdout, "\n");
   }
   return retval;
}

/***********************************************************************/
static int AllocAndFillDatabuf (
   FILE *fpi, long len, struct rhmidi_eventdata *p_evtdat)
{
   size_t red;

   free(p_evtdat->EventData);
   p_evtdat->EventData = malloc(len);
   if (!p_evtdat->EventData)
   {
      return 3;
   }
   if ((red = fread(p_evtdat->EventData, 1, len, fpi)) != (unsigned long) len)
   {
      return 4;
   }
   return 0;
}
